package pl.elzab.virtualcomsample;

import android.app.Activity;
import android.content.Context;
import android.util.Log;
import com.hoho.android.usbserial.driver.UsbSerialPort;
import com.hoho.android.usbserial.util.HexDump;
import com.hoho.android.usbserial.util.SerialInputOutputManager;
import java.io.IOException;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import pl.elzab.virtualcomsample.dataobject.IngoingResponse;
import pl.elzab.virtualcomsample.dataobject.Measurement;
import pl.elzab.virtualcomsample.dataobject.StatusInfo;
import pl.elzab.virtualcomsample.dataobject.StxErrorMessage;
import pl.elzab.virtualcomsample.dataobject.StxReceiptPosition;
import pl.elzab.virtualcomsample.protocol.AngelProtocol;
import pl.elzab.virtualcomsample.protocol.ElzabProtocol;
import pl.elzab.virtualcomsample.protocol.StxProtocol;

/**
 * Class for exchanging data with USB device. Supported protocols: Elzab, Angel, STX.
 * @author ElzabSoft
 * @version 1.01
 */
public class Communication {
    /** tag for logging */
    public static final String TAG = "ElzabUSB";
    /** USB I/O manager */
    protected SerialInputOutputManager mSerialIOManager;
    /** USB I/O listener */
    protected SerialInputOutputManager.Listener mListener;
    /** response timeout of USB device */
    protected final int mTimeoutMillis = 500;
    /** amount of loops in which we wait for response */
    protected final int mWaitLoopsAmount = 20;
    /** Executor used for listening the port. */
    protected final ExecutorService mExecutor = Executors.newSingleThreadExecutor();
    /** object that interprets ingoing responses */
    protected IngoingResponseHandler mIngoingResponseHandler = new IngoingResponseHandler();
    /** application context */
    protected Context mContext;
    /** STX text charset */
    protected String mStxCharset;
    /** received responses */
    LinkedList<IngoingResponse> mResponses;
    /** helper flag indicating that there is already started dialog */
    private boolean mPendingDialog = false;

    /** basic constructor
     * @param activity activity that provides ui thread to run listener on
     */
    public Communication(final Activity activity) {
        mResponses = new LinkedList<IngoingResponse>();
        mContext = activity;
        if( mListener == null) {
            Log.d(TAG, "Initializing mListener...");
            mListener = new SerialInputOutputManager.Listener() {
                @Override
                public void onNewData(final byte[] data) {
                    activity.runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            updateReceivedData(data);
                        }
                    });
                }
                @Override
                public void onRunError(Exception e) {
                    Log.d(TAG, e.getMessage());
                }
            };
        } else
            Log.d(TAG, "Communication(): mListener was already initialized...");
    }

    /** runs dialog for printing fiscal receipt
     * @param argPositions fiscal receipt positions
     */
    public void stxPrintReceipt(final List<StxReceiptPosition> argPositions) {
        if(!mPendingDialog) {
            Thread t = new Thread(new Runnable() {
                public void run() {
                    mPendingDialog = true;
                    UsbSerialPort port = USBConnection.getConnection(mContext);
                    IngoingResponse response;
                    float totalPrice = 0.0f;
                    try {
                        if(port != null) {
                            restartListener(port);
                            mIngoingResponseHandler.ClearMessageBuffer();
                            mResponses.clear();
                            port.write(StxProtocol.StxGetDeviceInfo(), mTimeoutMillis);
                            response = waitForResponse(IngoingResponse.ResponseType.STX_DEVICE_PRESENCE, IngoingResponse.ResponseType.STX_ERROR);
                            if(StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_DEVICE_PRESENCE)) {
                                if(!StxProtocol.IsDeviceAPrinter(response.GetResponseBytes())) {
                                    StatusInfo statusInfo = new StatusInfo("Urządzenie nie jest drukarką.", StatusInfo.MessageStatus.WARNING);
                                    Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
                                } else {
                                    mIngoingResponseHandler.ClearMessageBuffer();
                                    mResponses.clear();
                                    port.write(StxProtocol.StxGetCoding(), mTimeoutMillis);
                                    response = waitForResponse(IngoingResponse.ResponseType.STX_CODING_INFO, IngoingResponse.ResponseType.STX_ERROR);
                                    if(StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_CODING_INFO)) {
                                        mStxCharset = StxProtocol.ParseCodingInfo(response.GetResponseBytes());
                                        port.write(StxProtocol.StxReceiptOpen(), mTimeoutMillis);
                                        response = waitForResponse(IngoingResponse.ResponseType.STX_EMPTY_FRAME, IngoingResponse.ResponseType.STX_ERROR);
                                        if(StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_EMPTY_FRAME)) {
                                            for(StxReceiptPosition position : argPositions) {
                                                totalPrice += position.GetPrice() * position.GetQuantity();
                                                port.write(StxProtocol.StxReceiptAddPosition(position), mTimeoutMillis);
                                                response = waitForResponse(IngoingResponse.ResponseType.STX_EMPTY_FRAME, IngoingResponse.ResponseType.STX_ERROR);
                                                if(!StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_EMPTY_FRAME))
                                                    return;
                                            }
                                            port.write(StxProtocol.StxReceiptClose(totalPrice), mTimeoutMillis);
                                            response = waitForResponse(IngoingResponse.ResponseType.STX_EMPTY_FRAME, IngoingResponse.ResponseType.STX_ERROR);
                                            StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_EMPTY_FRAME);
                                        }
                                    }
                                }
                            } else {
                                StatusInfo statusInfo = new StatusInfo("Nie wykryto protokołu STX.", StatusInfo.MessageStatus.WARNING);
                                Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
                            }
                        }
                    } catch(IOException | InterruptedException e) {
                        USBConnection.closeConnection(mContext);
                    }
                    mPendingDialog = false;
                }
            });
            t.start();
        } else {
            StatusInfo statusInfo = new StatusInfo("Trwa dialog.", StatusInfo.MessageStatus.ERROR);
            Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
            Log.d(TAG, "Unable to start dialog. There is a pending dialog.");
        }
    }

    /** runs dialog for getting device's unique number */
    public void stxGetUniqueNumber() {
        if(!mPendingDialog) {
            Thread t = new Thread(new Runnable() {
                public void run() {
                    mPendingDialog = true;
                    UsbSerialPort port = USBConnection.getConnection(mContext);
                    IngoingResponse response;
                    try {
                        if(port != null) {
                            restartListener(port);
                            mIngoingResponseHandler.ClearMessageBuffer();
                            mResponses.clear();
                            port.write(StxProtocol.StxGetDeviceInfo(), mTimeoutMillis);
                            response = waitForResponse(IngoingResponse.ResponseType.STX_DEVICE_PRESENCE, IngoingResponse.ResponseType.STX_ERROR);
                            if (StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_DEVICE_PRESENCE)) {
                                mIngoingResponseHandler.ClearMessageBuffer();
                                mResponses.clear();
                                port.write(StxProtocol.StxGetUniqueNr(), mTimeoutMillis);
                                response = waitForResponse(IngoingResponse.ResponseType.STX_UNIQUE_NR, IngoingResponse.ResponseType.STX_ERROR);
                                if(StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_UNIQUE_NR)) {
                                    String uniqueNumber = StxProtocol.ParseUniqueNr(response.GetResponseBytes());
                                    Util.sendString(mContext, uniqueNumber, "message-for-activity", "stx_unique_number");
                                }
                            } else {
                                StatusInfo statusInfo = new StatusInfo("Nie wykryto protokołu STX.", StatusInfo.MessageStatus.WARNING);
                                Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
                            }

                        }
                    } catch(IOException | InterruptedException e) {
                        USBConnection.closeConnection(mContext);
                    }
                    mPendingDialog = false;
                }
            });
            t.start();
        } else {
            StatusInfo statusInfo = new StatusInfo("Trwa dialog.", StatusInfo.MessageStatus.ERROR);
            Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
            Log.d(TAG, "Unable to start dialog. There is a pending dialog.");
        }
    }

    /**
     * runs dialog for getting elzab weight measurement
     * @param scalesNumber scales' number
     * @param measurementFormat measurement format (basic or extended)
     */
    public void elzabWeightMeasurement(final ElzabProtocol.ScalesNumber scalesNumber,final ElzabProtocol.MeasurementFormat measurementFormat) {
        if(!mPendingDialog) {
            Thread t = new Thread(new Runnable() {
                public void run() {
                    mPendingDialog = true;
                    Log.d(TAG, "Method elzabWeightMeasurement invokes method USBConnection.getConnection(Context)");
                    UsbSerialPort port = USBConnection.getConnection(mContext);
                    IngoingResponse response;
                    Measurement measurement;
                    try {
                        if(port != null) {
                            restartListener(port);
                            mIngoingResponseHandler.ClearMessageBuffer();
                            mResponses.clear();
                            port.write(StxProtocol.StxGetDeviceInfo(), mTimeoutMillis);
                            response = waitForResponse(IngoingResponse.ResponseType.STX_DEVICE_PRESENCE);
                            if(StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_DEVICE_PRESENCE)) {
                                StatusInfo statusInfo = new StatusInfo("Urządzenie nie jest wagą.", StatusInfo.MessageStatus.WARNING);
                                Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
                            } else {
                                mIngoingResponseHandler.ClearMessageBuffer();
                                mResponses.clear();
                                port.write(ElzabProtocol.ElzabScalesPresence(scalesNumber), mTimeoutMillis);
                                response = waitForResponse(IngoingResponse.ResponseType.ELZAB_PROTOCOL_PRESENCE);
                                if(response != null && response.GetResponseType() == IngoingResponse.ResponseType.ELZAB_PROTOCOL_PRESENCE) {
                                    port.write(ElzabProtocol.ElzabMeasurement(scalesNumber, measurementFormat), mTimeoutMillis);
                                    response = waitForResponse(IngoingResponse.ResponseType.ELZAB_BASIC_MEASUREMENT, IngoingResponse.ResponseType.ELZAB_EXTENDED_MEASUREMENT);
                                    if( response != null) {
                                        if (response.GetResponseType() == IngoingResponse.ResponseType.ELZAB_BASIC_MEASUREMENT) {
                                            measurement = ElzabProtocol.ParseElzabBasicMeasurement(response.GetResponseBytes());
                                            Util.sendParcelable(mContext, measurement, "message-for-activity", "weight_measurement");
                                        } else if (response.GetResponseType() == IngoingResponse.ResponseType.ELZAB_EXTENDED_MEASUREMENT) {
                                            measurement = ElzabProtocol.ParseElzabExtendedMeasurement(response.GetResponseBytes());
                                            Util.sendParcelable(mContext, measurement, "message-for-activity", "weight_measurement");
                                        }
                                    }
                                } else {
                                    port.write(AngelProtocol.AngelPresence(), mTimeoutMillis);
                                    response = waitForResponse(IngoingResponse.ResponseType.ANGEL_PROTOCOL_PRESENCE);
                                    if( response != null && response.GetResponseType() == IngoingResponse.ResponseType.ANGEL_PROTOCOL_PRESENCE) {
                                        StatusInfo statusInfo = new StatusInfo("Urządzenie pracuje w protokole Angel.", StatusInfo.MessageStatus.WARNING);
                                        Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
                                    }
                                }
                            }
                        }
                    } catch(IOException | InterruptedException e) {
                        USBConnection.closeConnection(mContext);
                    }
                    mPendingDialog = false;
                }
            });
            t.start();
        } else {
            StatusInfo statusInfo = new StatusInfo("Trwa dialog.", StatusInfo.MessageStatus.ERROR);
            Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
            Log.d(TAG, "Unable to start dialog. There is a pending dialog.");
        }
    }

    /** runs dialog for getting angel weight measurement */
    public void angelWeightMeasurement() {
        if(!mPendingDialog) {
            Thread t = new Thread(new Runnable() {
                public void run() {
                    mPendingDialog = true;
                    Log.d(TAG, "Method angelWeightMeasurement() invokes method USBConnection.getConnection(Context)");
                    UsbSerialPort port = USBConnection.getConnection(mContext);
                    IngoingResponse response = null;
                    Measurement measurement = null;
                    try {
                        if(port != null) {
                            restartListener(port);
                            mIngoingResponseHandler.ClearMessageBuffer();
                            mResponses.clear();
                            port.write(StxProtocol.StxGetDeviceInfo(), mTimeoutMillis);
                            response = waitForResponse(IngoingResponse.ResponseType.STX_DEVICE_PRESENCE);
                            if(StxProcessResponse(response, mStxCharset, port, IngoingResponse.ResponseType.STX_DEVICE_PRESENCE)) {
                                StatusInfo statusInfo = new StatusInfo("Urządzenie nie jest wagą.", StatusInfo.MessageStatus.WARNING);
                                Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
                            } else {
                                mIngoingResponseHandler.ClearMessageBuffer();
                                mResponses.clear();
                                port.write(AngelProtocol.AngelPresence(), mTimeoutMillis);
                                response = waitForResponse(IngoingResponse.ResponseType.ANGEL_PROTOCOL_PRESENCE);
                                if(response != null && response.GetResponseType() == IngoingResponse.ResponseType.ANGEL_PROTOCOL_PRESENCE) {
                                    port.write(AngelProtocol.AngelMeasurement(), mTimeoutMillis);
                                    response = waitForResponse(IngoingResponse.ResponseType.ANGEL_MEASUREMENT);
                                    if( response != null) {
                                        if (response.GetResponseType() == IngoingResponse.ResponseType.ANGEL_MEASUREMENT) {
                                            measurement = AngelProtocol.ParseAngelMeasurement(response.GetResponseBytes());
                                            Util.sendParcelable(mContext, measurement, "message-for-activity", "weight_measurement");
                                        }
                                    }
                                } else {
                                    port.write(ElzabProtocol.ElzabScalesPresence(ElzabProtocol.ScalesNumber.SCALES_0), mTimeoutMillis);
                                    response = waitForResponse(IngoingResponse.ResponseType.ELZAB_PROTOCOL_PRESENCE);
                                    if( response != null && response.GetResponseType() == IngoingResponse.ResponseType.ELZAB_PROTOCOL_PRESENCE) {
                                        StatusInfo statusInfo = new StatusInfo("Urządzenie pracuje w protokole Elzab.", StatusInfo.MessageStatus.WARNING);
                                        Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
                                    }
                                }
                            }
                        }
                    } catch(IOException | InterruptedException e) {
                        USBConnection.closeConnection(mContext);
                    }
                    mPendingDialog = false;
                }
            });
            t.start();
        } else {
            StatusInfo statusInfo = new StatusInfo("Trwa dialog.", StatusInfo.MessageStatus.ERROR);
            Util.sendParcelable(mContext, statusInfo, "message-for-activity", "status_info");
            Log.d(TAG, "Unable to start dialog. There is a pending dialog.");
        }
    }

    /**
     * helper method that processes ingoing response in STX dialog
     * @param response received response
     * @param charset charset
     * @param port USB port connection
     * @return whether to continue dialog
     * @throws IOException
     * @throws InterruptedException
     */
    private boolean StxProcessResponse(IngoingResponse response, String charset, UsbSerialPort port, IngoingResponse.ResponseType requiredResponseType) throws IOException, InterruptedException{
        if(response != null) {
            if(response.GetResponseType() != requiredResponseType ) {
                if(response.GetResponseType() == IngoingResponse.ResponseType.STX_ERROR) {
                    StxErrorMessage errorMessage = stxGetErrorMessage(port, StxProtocol.ParseStxErrorCode(response.GetResponseBytes()), mStxCharset);
                    if (errorMessage != null)
                        Util.sendParcelable(mContext, errorMessage, "message-for-activity", "stx_error_message");
                    return false;
                }
                Log.d(TAG, "Incorrect response. Required response: " + IngoingResponse.ResponseType.STX_EMPTY_FRAME.toString());
                return false;
            }
            return true;
        }
        Log.d(TAG, "StxProcessResponse: Error: null response object!");
        return false;
    }

    /**
     * method used to retrieve stx error message from device
     * @param port USB port connection
     * @param stxErrorCode error code from STX device
     * @param charset charset using by STX device
     * @return stx error message data
     * @throws IOException
     * @throws InterruptedException
     */
    private StxErrorMessage stxGetErrorMessage(UsbSerialPort port, String stxErrorCode, String charset) throws IOException, InterruptedException{
        port.write(StxProtocol.StxGetErrorDescription(stxErrorCode), mTimeoutMillis);
        IngoingResponse response = waitForResponse(IngoingResponse.ResponseType.STX_ERROR_DESCRIPTION);
        if(response != null && response.GetResponseType() == IngoingResponse.ResponseType.STX_ERROR_DESCRIPTION) {
            return StxProtocol.ParseStxError(response.GetResponseBytes(), charset);
        }
        return null;
    }

    /**
     * waits for specified response
     * @param requiredResponseType required response type
     * @return received response or null in case of not getting proper response
     */
    private IngoingResponse waitForResponse(IngoingResponse.ResponseType... requiredResponseType) throws InterruptedException {
        IngoingResponse ingoingResponse = null;
        long delayTime = mTimeoutMillis / mWaitLoopsAmount;
        for(int i=0, stopTimeout = 0; i<mWaitLoopsAmount && stopTimeout == 0; i++) {
            while(mResponses.size() > 0) {
                Log.d(TAG, "Thread: Got response! " + mResponses.getFirst().GetResponseType().toString());
                if(Arrays.asList(requiredResponseType).contains(mResponses.getFirst().GetResponseType())) {
                    stopTimeout = 1;
                    ingoingResponse = mResponses.getFirst();
                    mResponses.removeFirst();
                    break;
                }
                mResponses.removeFirst();
            }
            Thread.sleep(delayTime);
        }
        //if(ingoingResponse == null) {
        //    Util.sendParcelable(mContext, new StatusInfo("Timeout odpowiedzi.", StatusInfo.MessageStatus.ERROR), "message-for-activity", "status_info");
        //}
        return ingoingResponse;
    }

    /**
     * Stops USB I/O listener.
     */
    public synchronized void stopListener() {
        if (mSerialIOManager != null) {
            mSerialIOManager.stop();
            mSerialIOManager = null;
            Log.d(TAG, "Listener has been stopped.");
        } else {
            Log.d(TAG, "Listener has already been stopped!");
        }
    }

    /**
     * Starts USB I/O listener.
     * (run it after the port is opened)
     * @param port USB Serial port instance
     */
    public void startListener(UsbSerialPort port) {
        if (port == null || mListener == null || mSerialIOManager != null ) {
            Log.d(TAG, "Unable to start listener!");
            return;
        }
        stopListener();
        mSerialIOManager = new SerialInputOutputManager(port, mListener);
        mExecutor.submit(mSerialIOManager);
        Log.d(TAG, "Listener has been started.");
    }

    /**
     * Restarts USB I/O listener.
     * @param port USB Serial port instance
     */
    public void restartListener(UsbSerialPort port) {
        stopListener();
        startListener(port);
    }

    /** receives raw bytes from serial port and processes it
     * @param data raw byte data
     */
    //protected synchronized void updateReceivedData(byte[] data) {
    protected synchronized void updateReceivedData(byte[] data) {
        Log.d(TAG, "Received data: " + HexDump.toHexString(data));
        IngoingResponse response = mIngoingResponseHandler.ProcessSerialOutput(data); // process raw bytes
        IngoingResponse.ResponseType responseType = response.GetResponseType();

        if( responseType != IngoingResponse.ResponseType.UNKNOWN) {
            if(mPendingDialog) {
                Log.d(TAG, "Adding response.");
                mResponses.addLast(response);
            } else { // unexpected data
                processUnexpectedData(response);
            }
        }
    }

    /**
     * processes unexpected data (for example weight measurement initiated on scales side)
     * @param response
     */
    private void processUnexpectedData(IngoingResponse response) {
        Log.d(TAG, "Processing unexpected data.");
        Measurement measurement = null;
        if(response.GetResponseType() == IngoingResponse.ResponseType.ELZAB_BASIC_MEASUREMENT) {
            measurement = ElzabProtocol.ParseElzabBasicMeasurement(response.GetResponseBytes());
            Util.sendParcelable(mContext, measurement, "message-for-activity", "weight_measurement");
        } else if(response.GetResponseType() == IngoingResponse.ResponseType.ELZAB_EXTENDED_MEASUREMENT) {
            measurement = ElzabProtocol.ParseElzabExtendedMeasurement(response.GetResponseBytes());
            Util.sendParcelable(mContext, measurement, "message-for-activity", "weight_measurement");
        } else if(response.GetResponseType() == IngoingResponse.ResponseType.ANGEL_MEASUREMENT) {
            measurement = AngelProtocol.ParseAngelMeasurement(response.GetResponseBytes());
            Util.sendParcelable(mContext, measurement, "message-for-activity", "weight_measurement");
        }
    }
}
