package pl.elzab.virtualcomsample.protocol;

import android.util.Log;

import pl.elzab.virtualcomsample.dataobject.Measurement;

/**
 * Class used to generate or parse Angel protocol frames.
 * @author ElzabSoft
 * @version 1.0
 */
public class AngelProtocol {
    /** tag for logging */
    public static final String TAG = "ElzabUSB";
    /**
     * check wheter the frame received by device indicates that it's angel protocol device
     * @param frame received frame
     * @return true if it's angel protocol device
     */
    public static boolean IsAngelPresence(byte[] frame) {
        return frame.length == 1 && frame[0] == 0x06;
    }

    /**
     * check whether the frame received by device indicates that it's angel measurement frame
     * @param frame received frame
     * @return true if it's angel measurement frame
     */
    public static boolean IsAngelMeasurement(byte[] frame) {
        return frame.length == 15 && (frame[2] == 0x53 || frame[2] == 0x55) && frame[4] == 0x20 && frame[14] == 0x04;
    }

    /**
     * generates byte command to check if device is set to Angel protocol
     * @return byte command
     */
    public static byte[] AngelPresence() {
        return new byte[] {0x05};
    }

    /**
     * generates byte command to get weight measurement in Angel protocol
     * @return byte command
     */
    public static byte[] AngelMeasurement() {
        return new byte[] {0x11};
    }

    /**
     * parses Measurement from Angel protocol frame
     * @param data frame bytes
     * @return Measurement object
     */
    public static Measurement ParseAngelMeasurement(byte[] data) {
        Measurement measurement = null;
        Log.d(TAG, "Parsing Angel measurement.");
        if (data.length == 15) {
            measurement = new Measurement();
            byte[] array = new byte[15];
            for (int i = 0; i < 15; i++)
                array[i] = data[i];
            StringBuilder stringbuilder = new StringBuilder();
            stringbuilder.append(new String(array, 5, 5));
            try {
                measurement.SetWeight(Double.parseDouble(stringbuilder.toString()));
                if (data[3] == 0x2D) // minus sign, so negate the value
                    measurement.SetWeight(measurement.GetWeight() * -1.0);
            } catch (Exception e) {
                measurement.SetWeight(0.0);
            }
            measurement.SetUnit(data[10] == 0x6B ? Measurement.Unit.KILOGRAM : Measurement.Unit.POUND);
            measurement.SetOverflow(data[3] == 0x46);
            measurement.SetStability(data[2] == 0x53 ? Measurement.Stability.STABLE : Measurement.Stability.UNSTABLE);
        }
        Log.d(TAG, "Parsing Angel measurement finished.");
        return measurement;
    }
}
