package pl.elzab.virtualcomsample.protocol;

import android.util.Log;
import java.io.UnsupportedEncodingException;
import pl.elzab.virtualcomsample.dataobject.StxErrorMessage;
import pl.elzab.virtualcomsample.Util;
import pl.elzab.virtualcomsample.dataobject.StxReceiptPosition;

/**
 * Class used to generate or parse STX protocol frames.
 * @author ElzabSoft
 * @version 1.0
 */
public class StxProtocol {
    /** tag for logging */
    public static final String TAG = "ElzabUSB";
    /** minimal lenght of stx frame */
    public static final int STX_FRAME_MIN_LENGTH = 7;
    /** minimal length of unique number */
    public static final int UNIQUE_NR_MIN_LENGTH = 1;
    /** maximal length of unique number */
    public static final int UNIQUE_NR_MAX_LENGTH = 14;
    /** maximal length of stx error description */
    public static final int STX_ERROR_DESC_MAX_LENGTH = 20;
    /** stx crc bytes length */
    public static final int STX_CRC_LENGTH = 4;
    /** maximal length of stx error code */
    public static final int STX_ERORRCODE_MAX_LENGTH = 4;
    /** name of elzab zeta device that appears in stx frame */
    public static final String ELZAB_ZETA_NAME = "ELZAB ZETA";
    /** name of elzab d10 device that appears in stx frame */
    public static final String ELZAB_D10_NAME = "ELZAB D10";

    /**
     * generates crc from byte array
     * @param int_input
     * @return
     */
    static public int crc_bytes(byte[] int_input)
    {
        int len = int_input.length;
        int int_crc_byte_a, int_crc_byte_b;
        byte[] int_array = new byte[100];
        int_array = int_input;
        int int_crc = 0x0;
        int int_lsb;
        for (int int_i = 0; int_i < len; int_i++)
        {
            int_crc = int_crc ^ int_array[int_i];
            for (int int_j = 0; int_j < 8; int_j++)
            {
                int_lsb = int_crc & 0x0001;  // Mask of LSB
                int_crc = int_crc >> 1;
                int_crc = int_crc & 0x7FFF;
                if (int_lsb == 1) int_crc = int_crc ^ 0xA001;
            }
        }
        int_crc_byte_a = int_crc & 0x00FF;
        int_crc_byte_b = (int_crc >> 8) & 0x00FF;
        return int_crc_byte_b * 0x100 + int_crc_byte_a;
    }

    /**
     * generates crc byte array based on integer value
     * @param crc16 crc integer value
     * @return crc field for stx frame
     */
    static public byte[] GetCrcField(int crc16){
        String text = Integer.toHexString(crc16);
        byte[] field = new byte[4];
        for(int j=0; j<4; j++)
            field[j] = (byte)text.charAt(j);
        return field;
    }

    /**
     * check whether the data field of stx frame contains charset info
     * @param dataField stx data field
     * @return true if it's charset info
     */
    static public boolean IsCodingInfo(byte[] dataField) {
        boolean result = true;
        if( dataField.length > 3 && dataField[0] == 0x1d && dataField[1] == 'w' && dataField[2] == 'A') {
            for (int i = 3; i < dataField.length; i++)
                if (!IsPrintableChar(dataField[i])) {
                    result = false;
                    break;
                }
        } else
            result = false;
        return result;
    }

    /**
     * parses charset symbol from stx charset info frame
     * @param stxFrame stx frame bytes
     * @return charset symbol
     */
    public static String ParseCodingInfo(byte[] stxFrame) {
        byte[] dataField = GetDataField(stxFrame);
        int length = dataField.length - 3;
        String result = null;
        if( length > 0) {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.append(new String(dataField, 3, length));
            result = stringBuilder.toString();
        }
        return result;
    }

    /**
     * parses data field from stx frame
     * @param stxFrame stx frame bytes
     * @return stx data field
     */
    static public byte[] GetDataField(byte[] stxFrame) {
        if( stxFrame.length < STX_FRAME_MIN_LENGTH)
            return null;
        int dataFieldPosition = -1;
        int crcPosition = stxFrame.length - STX_CRC_LENGTH - 1;
        for(int i=0; i<crcPosition; i++)
            if(stxFrame[i] == 0x1C) {
                dataFieldPosition = i+1;
                break;
            }
        if( dataFieldPosition > -1) {
            byte[] dataField = new byte[crcPosition-dataFieldPosition];
            for(int i=0; i<dataField.length; i++)
                dataField[i] = stxFrame[i+dataFieldPosition];
            return dataField;
        }
        return null;
    }

    /**
     * parses header field from stx frame
     * @param stxFrame stx frame bytes
     * @return stx header field
     */
    static public byte[] GetHeaderField(byte[] stxFrame) {
        int length = stxFrame.length;
        if( length < STX_FRAME_MIN_LENGTH)
            return null;
        int headerFieldLength = 0;
        for(int i=1; i<length; i++) {
            if(stxFrame[i] == 0x1C)
                break;
            headerFieldLength++;
        }
        byte[] headerField = new byte[headerFieldLength];
        for(int i=0; i<headerFieldLength; i++)
            headerField[i] = stxFrame[i+1];
        return headerField;
    }

    /**
     * checks whether the received frame is stx frame
     * @param frame received data frame
     * @return true if it's stx frame
     */
    static public boolean IsStxFrame(byte[] frame) {
        return frame.length >= StxProtocol.STX_FRAME_MIN_LENGTH && frame[0] == 0x02 && frame[frame.length-1] == 0x03;
    }

    /**
     * check whether the byte is printable character
     * @param c character
     * @return true if it's printable character
     */
    static public boolean IsPrintableChar(byte c) {
        return c >= ' ';
    }

    /**
     * check whether the stx data field contains devices' unique number
     * @param dataField stx data field bytes
     * @return true if data field contains unique number
     */
    static public boolean IsUniqueNumber(byte[] dataField) {
        if( dataField.length < 2 + UNIQUE_NR_MIN_LENGTH || dataField.length > 2 + UNIQUE_NR_MAX_LENGTH )
            return false;
        if( dataField[0] == 0x1d && dataField[1] == 'B') {
            int i = dataField.length - 1;
            while(i > 1) {
                if(!IsPrintableChar(dataField[i]))
                    return false;
                i--;
            }
            return true;
        }
        return false;
    }

    /**
     * checks wheter the stx data field contains error code
     * @param dataField stx data field bytes
     * @return true if it contains error code
     */
    static public boolean IsError(byte[] dataField) {
        if(dataField.length > 0 && dataField.length < STX_ERORRCODE_MAX_LENGTH) {
            for(byte b : dataField)
                if(!IsDigit(b))
                    return false;
        } else
            return false;
        return true;
    }

    /**
     * check whether the byte is digit
     * @param b input byte
     * @return true if the byte is digit
     */
    static public boolean IsDigit(byte b) {
        return b > '/' && b < ':';
    }

    /**
     * checks whether the stx data field contains stx error description
     * @param dataField stx data field bytes
     * @return true if it contains stx error description
     */
    static public boolean IsErrorDescription(byte[] dataField) {
        int length = dataField.length;
        if(length >= 4 && dataField[0] == 0x1D && dataField[1] == 'T') {
            int errorDescLength = 0;
            for(int i=2; i<length - 1; i++) {
                if(dataField[i] == 0x1D && dataField[i+1] == 'U') {
                    break;
                }
                else
                    errorDescLength++;
            }
            if(errorDescLength > STX_ERROR_DESC_MAX_LENGTH)
                return false;
            return true;
        }
        return false;
    }

    /**
     * checks if the stx frame is an empty frame
     * @param dataField stx data field bytes
     * @return true if it's an empty frame
     */
    static public boolean IsEmptyFrame(byte[] dataField) {
        return dataField.length < 1;
    }

    /**
     * parses stx error code from stx frame
     * @param stxFrame stx frame bytes
     * @return stx error code
     */
    static public String ParseStxErrorCode(byte[] stxFrame) {
        byte[] dataField = GetDataField(stxFrame);
        Log.d(TAG, "Parsing STX Error code from error frame.");
        StringBuilder stringBuilder = new StringBuilder();
        stringBuilder.append(new String(dataField, 0, dataField.length));
        Log.d(TAG, "ParseStxErrorCode: returning " + stringBuilder.toString());
        return stringBuilder.toString();
    }

    /**
     * parses stx error code from error description frame
     * @param dataField stx data field bytes
     * @return stx error code
     */
    static public String ParseStxErrorCodeFromErrorDesc(byte[] dataField) {
        Log.d(TAG, "Parsing STX Error code from error description frame.");
        int length = dataField.length;
        int errorCodeIndex = -1;
        int errorCodeLength = 0;
        for(int i=0; i<length-2; i++)
            if(dataField[i] == 0x1D && dataField[i+1] == 'T') {
                errorCodeIndex = i+2;
                break;
            }
        for(int i=errorCodeIndex; i<length; i++)
            if(dataField[i] == 0x1D)
                break;
            else
                errorCodeLength++;
        if(errorCodeIndex == -1 || errorCodeLength < 1)
            return null;
        StringBuilder stringBuilder = new StringBuilder();
        stringBuilder.append(new String(dataField, errorCodeIndex, errorCodeLength));
        Log.d(TAG, "ParseStxErrorCodeFromErrorDesc: returning " + stringBuilder.toString());
        return stringBuilder.toString();
    }

    /**
     * parses stx error text from stx data field
     * @param dataField stx data field bytes
     * @param charset charset used by stx device
     * @return stx error text
     */
    static public String ParseStxErrorText(byte[] dataField, String charset) {
        Log.d(TAG, "Parsing STX Error text.");
        int length = dataField.length;
        int errorTextIndex = -1;
        int errorTextLength = 0;
        for(int i=0; i<length-2; i++)
            if(dataField[i] == 0x1D && dataField[i+1] == 'U') {
                errorTextIndex = i+2;
                break;
            }
        errorTextLength = length - errorTextIndex;
        if(errorTextIndex == -1 || errorTextLength < 1)
            return null;
        Log.d(TAG, "ParseStxErrorText: charset: " + charset);
        return DecodeText(dataField, errorTextIndex, errorTextLength, charset);
    }

    /**
     * decodes text contained in byte array
     * @param byteArray text to decode
     * @param arrayStart start of array
     * @param arrayLength array fragment length
     * @param charset charset used by stx device
     * @return decoded text
     */
    static public String DecodeText(byte[] byteArray, int arrayStart, int arrayLength, String charset) {
        String result = "";
        if( charset.equals("CP852"))
            charset = "IBM852";
        try {
            result = new String(byteArray, arrayStart, arrayLength, charset);
        } catch (UnsupportedEncodingException e) {
            Log.d(TAG, "DecodeText: " + e.getMessage());
            result = new String(byteArray, arrayStart, arrayLength);
        }
        return result;
    }

    /**
     * parses stx error data from stx frame
     * @param stxFrame stx frame bytes
     * @param charset charset used by stx device
     * @return stx error message data
     */
    static public StxErrorMessage ParseStxError(byte[] stxFrame, String charset) {
        Log.d(TAG, "Parsing STX error from frame.");
        byte[] dataField = GetDataField(stxFrame);
        StxErrorMessage error = new StxErrorMessage();
        if( dataField != null ) {
            //error.SetErrorDesc(ParseStxErrorDesc(dataField));
            error.SetErrorText(ParseStxErrorText(dataField, charset));
            error.SetErrorCode(ParseStxErrorCodeFromErrorDesc(dataField));
        }
        return error;
    }

    /**
     * parses devices' unique number from stx frame
     * @param stxFrame stx frame bytes
     * @return devices' unique number
     */
    static public String ParseUniqueNr(byte[] stxFrame) {
        Log.d(TAG, "Parsing STX unique number.");
        byte[] dataField = GetDataField(stxFrame);
        if( dataField == null)
            return null;
        StringBuilder stringbuilder = new StringBuilder();
        stringbuilder.append(new String(dataField, 2, dataField.length-2));
        Log.d(TAG, "Parsing STX unique number finished.");
        return stringbuilder.toString();
    }

    /**
     * chechs whether the stx data field contains information about device
     * @param dataField stx data field bytes
     * @return true if it contains device information
     */
    static public boolean IsDeviceInfo(byte[] dataField) {
        int length = dataField.length;
        boolean result = false;
        if(length > 4 && dataField[0] == 0x1d && dataField[1] == 'B') {
            for(int i=2; i<length; i++) {
                if(dataField[i] == 0x1d) {
                    result = dataField[i+1] == 'C';
                    break;
                }
            }
        }
        return result;
    }

    /**
     * parses device name from stx data field
     * @param dataField stx data field bytes
     * @return device name
     */
    static public String ParseDeviceName(byte[] dataField) {
        String result = "";
        int length = dataField.length;
        int start = 2;
        int end = -1;
        for(int i=2; i<length-2; i++) {
            if(dataField[i+1] == 0x1d && dataField[i+2] == 'C') {
                end = i;
                break;
            }
        }
        if(end > 2)
            result = new String(dataField, start, end - start + 1);
        return result;
    }

    /**
     * check whether the device is a printer (elzab zeta/elzab d10)
     * @param stxFrame stx frame bytes
     * @return true if it's a printer
     */
    static public boolean IsDeviceAPrinter(byte[] stxFrame) {
        String deviceName = ParseDeviceName(GetDataField(stxFrame));
        boolean result = deviceName.equals(ELZAB_D10_NAME) || deviceName.equals(ELZAB_ZETA_NAME);
        if( result )
            Log.d(TAG, "IsDeviceAPrinter: device is a printer! (" + deviceName + ")");
        else
            Log.d(TAG, "IsDeviceAPrinter: device is NOT a printer! (" + deviceName + ")");
        return result;
    }

    /**
     * generates byte command to get device's unique number
     * @return byte command
     */
    public static byte[] StxGetUniqueNr() {
        byte[] stx = new byte[] {0x02};
        byte[] headerField = new byte[] {0x30, 0x1D, 0x41, 0x4C, 0x4C, 0x1D, 0x31, 0x1D, 0x31, 0x1D};
        byte[] fileSeparator = new byte[] {0x1C};
        byte[] dataField = new byte[] {'F', 'N', 'R'};
        byte[] headerAndData = Util.ConcatenateByteArrays(headerField, fileSeparator, dataField);
        int crcSum = StxProtocol.crc_bytes(headerAndData);
        byte[] crcField = StxProtocol.GetCrcField(crcSum);
        byte[] etx = new byte[] {0x03};
        return Util.ConcatenateByteArrays(stx, headerAndData, crcField, etx);
    }

    /**
     * generates byte command to get device information
     * @return byte command
     */
    public static byte[] StxGetDeviceInfo() {
        byte[] stx = new byte[] {0x02};
        byte[] headerField = new byte[] {0x30, 0x1D, 0x41, 0x4C, 0x4C, 0x1D, 0x31, 0x1D, 0x31, 0x1D};
        byte[] fileSeparator = new byte[] {0x1C};
        byte[] dataField = new byte[] {'X', 'T', 'R'};
        byte[] headerAndData = Util.ConcatenateByteArrays(headerField, fileSeparator, dataField);
        int crcSum = StxProtocol.crc_bytes(headerAndData);
        byte[] crcField = StxProtocol.GetCrcField(crcSum);
        byte[] etx = new byte[] {0x03};
        return Util.ConcatenateByteArrays(stx, headerAndData, crcField, etx);
    }

    /**
     * generates byte command to get information about text coding used by device
     * @return byte command
     */
    public static byte[] StxGetCoding() {
        byte[] stx = new byte[] {0x02};
        byte[] headerField = new byte[] {0x30, 0x1D, 0x41, 0x4C, 0x4C, 0x1D, 0x31, 0x1D, 0x31, 0x1D};
        byte[] fileSeparator = new byte[] {0x1C};
        byte[] dataField = new byte[] {'X', 'H', 'R'};
        byte[] headerAndData = Util.ConcatenateByteArrays(headerField, fileSeparator, dataField, new byte[] {0x1d, 'w'});
        int crcSum = StxProtocol.crc_bytes(headerAndData);
        byte[] crcField = StxProtocol.GetCrcField(crcSum);
        byte[] etx = new byte[] {0x03};
        return Util.ConcatenateByteArrays(stx, headerAndData, crcField, etx);
    }

    /**
     * generates byte command to open receipt
     * @return byte command
     */
    public static byte[] StxReceiptOpen() {
        byte[] stx = new byte[] {0x02};
        byte[] headerField = new byte[] {0x34, 0x38};
        byte[] fileSeparator = new byte[] {0x1C};
        byte[] command = new byte[] {'L', 'O', 'R'};
        byte[] dataField = command;
        byte[] headerAndData = Util.ConcatenateByteArrays(headerField, fileSeparator, dataField);
        int crcSum = StxProtocol.crc_bytes(headerAndData);
        byte[] crcField = StxProtocol.GetCrcField(crcSum);
        byte[] etx = new byte[] {0x03};
        return Util.ConcatenateByteArrays(stx, headerAndData, crcField, etx);
    }

    /**
     * generates byte command to add receipt position
     * @param position receipt position
     * @return byte command
     */
    public static byte[] StxReceiptAddPosition(StxReceiptPosition position) {
        String strProductAmount = String.format("%.2f", position.GetQuantity());
        String strProductPrice = String.format("%.2f", position.GetPrice());
        String strTotalPrice = String.format("%.2f", position.GetQuantity() * position.GetPrice());
        strProductAmount.replace('.', ',');
        strProductPrice.replace('.', ',');
        strTotalPrice.replace('.', ',');
        Log.d(TAG, "strProductAmount: " + strProductAmount);
        Log.d(TAG, "strProductPrice: " + strProductPrice);
        Log.d(TAG, "strTotalPrice: " + strTotalPrice);
        byte[] stx = new byte[] {0x02};
        byte[] headerField = new byte[] {0x34, 0x39};
        byte[] fileSeparator = new byte[] {0x1C};
        byte[] command = new byte[] {'L', 'I', 'T'};
        byte[] name = Util.StringToByteArray(position.GetName());
        byte[] productAmount = Util.StringToByteArray(strProductAmount);
        byte[] productPrice = Util.StringToByteArray(strProductPrice);
        byte[] totalPrice = Util.StringToByteArray(strTotalPrice);
        byte[] unit = Util.StringToByteArray(position.GetQuantityUnit());
        byte[] dataField = Util.ConcatenateByteArrays(command, new byte[]{0x1d, 0x47, 0x41}, name,
                new byte[] {0x1e, 0x42, 0x41, 0x1e, 0x43}, productAmount, new byte[] {0x1e, 0x44}, productPrice, new byte[] {0x1e, 0x45}, totalPrice,
                new byte[] {0x1e, 0x47 }, unit);
        byte[] headerAndData = Util.ConcatenateByteArrays(headerField, fileSeparator, dataField);
        int crcSum = StxProtocol.crc_bytes(headerAndData);
        byte[] crcField = StxProtocol.GetCrcField(crcSum);
        byte[] etx = new byte[] {0x03};
        return Util.ConcatenateByteArrays(stx, headerAndData, crcField, etx);
    }

    /**
     * generates byte command to close receipt with given total price
     * @param fTotalPrice total price
     * @return byte command
     */
    public static byte[] StxReceiptClose(float fTotalPrice) {
        String strTotalPrice = String.format("%.2f", fTotalPrice);
        strTotalPrice.replace('.', ',');
        byte[] stx = new byte[] {0x02};
        byte[] headerField = new byte[] {0x35, 0x30};
        byte[] fileSeparator = new byte[] {0x1C};
        byte[] command = new byte[] {'L', 'C', 'L'};
        byte[] totalPrice = Util.StringToByteArray(strTotalPrice);
        byte[] dataField = Util.ConcatenateByteArrays(command, new byte[] {0x1d, 0x52, 0x41}, totalPrice );
        byte[] headerAndData = Util.ConcatenateByteArrays(headerField, fileSeparator, dataField);
        int crcSum = StxProtocol.crc_bytes(headerAndData);
        byte[] crcField = StxProtocol.GetCrcField(crcSum);
        byte[] etx = new byte[] {0x03};
        return Util.ConcatenateByteArrays(stx, headerAndData, crcField, etx);
    }

    /**
     * generates byte command to get STX error description based on error code
     * @param stxErrorCode STX error code
     * @return byte command
     */
    public static byte[] StxGetErrorDescription(String stxErrorCode) {
        byte[] stx = new byte[] {0x02};
        byte[] headerField = new byte[] {0x35, 0x30};
        byte[] fileSeparator = new byte[] {0x1C};
        byte[] command = new byte[] {'Y', 'E', 'R'};
        byte[] dataField = Util.ConcatenateByteArrays(command, new byte[] {0x1D, 'T'}, Util.StringToByteArray(stxErrorCode) );
        byte[] headerAndData = Util.ConcatenateByteArrays(headerField, fileSeparator, dataField);
        int crcSum = StxProtocol.crc_bytes(headerAndData);
        byte[] crcField = StxProtocol.GetCrcField(crcSum);
        byte[] etx = new byte[] {0x03};
        return Util.ConcatenateByteArrays(stx, headerAndData, crcField, etx);
    }
}
